%%%-------------------------------------------------------------------
%%% @author Michał Wrzeszcz
%%% @copyright (C) 2015 ACK CYFRONET AGH
%%% This software is released under the MIT license
%%% cited in 'LICENSE.txt'.
%%% @end
%%%-------------------------------------------------------------------
%%% @doc Information about tasks to be done.
%%% @end
%%%-------------------------------------------------------------------
-module(task_pool).
-author("Michał Wrzeszcz").
-behaviour(model_behaviour).

-include("elements/task_manager/task_manager.hrl").
-include("modules/datastore/datastore_internal_model.hrl").

%% model_behaviour callbacks
-export([save/1, get/1, list/0, list/1, list_failed/1, exists/1, delete/1, delete/2, update/2, update/3,
    create/1, create/2, model_init/0, 'after'/5, before/4]).
%% API
-export([count_tasks/3]).

%%%===================================================================
%%% model_behaviour callbacks
%%%===================================================================

%%--------------------------------------------------------------------
%% @doc
%% {@link model_behaviour} callback save/1. 
%% @end
%%--------------------------------------------------------------------
-spec save(datastore:document()) ->
    {ok, datastore:key()} | datastore:generic_error().
save(Document) ->
    datastore:save(?STORE_LEVEL, Document).

%%--------------------------------------------------------------------
%% @doc
%% {@link model_behaviour} callback update/2. 
%% @end
%%--------------------------------------------------------------------
-spec update(datastore:key(), Diff :: datastore:document_diff()) ->
    {ok, datastore:key()} | datastore:update_error().
update(Key, Diff) ->
    datastore:update(?STORE_LEVEL, ?MODULE, Key, Diff).

%%--------------------------------------------------------------------
%% @doc
%% Same as {@link model_behaviour} callback update/2 but allows
%% choice of task level.
%% @end
%%--------------------------------------------------------------------
-spec update(Level :: task_manager:level(), datastore:key(), Diff :: datastore:document_diff()) ->
    {ok, datastore:key()} | datastore:update_error().
update(?NON_LEVEL, _Key, _Diff) ->
    {ok, non};

update(Level, Key, Diff) ->
    datastore:update(task_to_db_level(Level), ?MODULE, Key, Diff).

%%--------------------------------------------------------------------
%% @doc
%% {@link model_behaviour} callback create/1. 
%% @end
%%--------------------------------------------------------------------
-spec create(datastore:document()) ->
    {ok, datastore:key()} | datastore:create_error().
create(Document) ->
    datastore:create(?STORE_LEVEL, Document).

%%--------------------------------------------------------------------
%% @doc
%% Same as {@link model_behaviour} callback create/1 but allows
%% choice of task level.
%% @end
%%--------------------------------------------------------------------
-spec create(Level :: task_manager:level(), datastore:document()) ->
    {ok, datastore:key()} | datastore:create_error().
create(?NON_LEVEL, _Document) ->
    {ok, non};

create(Level, Document) ->
    datastore:create(task_to_db_level(Level), Document).

%%--------------------------------------------------------------------
%% @doc
%% {@link model_behaviour} callback get/1.
%% @end
%%--------------------------------------------------------------------
-spec get(datastore:key()) -> {ok, datastore:document()} | datastore:get_error().
get(Key) ->
    datastore:get(?STORE_LEVEL, ?MODULE, Key).

%%--------------------------------------------------------------------
%% @doc
%% Returns list of all records.
%% @end
%%--------------------------------------------------------------------
-spec list() -> {ok, [datastore:document()]} | datastore:generic_error() | no_return().
list() ->
    datastore:list(?STORE_LEVEL, ?MODEL_NAME, ?GET_ALL, []).

%%--------------------------------------------------------------------
%% @doc
%% Returns list of all records.
%% @end
%%--------------------------------------------------------------------
-spec list(Level :: task_manager:level()) ->
    {ok, [datastore:document()]} | datastore:generic_error() | no_return().
list(?NON_LEVEL) ->
    {ok, []};

list(Level) ->
    datastore:list(task_to_db_level(Level), ?MODEL_NAME, ?GET_ALL, []).

%%--------------------------------------------------------------------
%% @doc
%% Returns list of tasks that failed.
%% @end
%%--------------------------------------------------------------------
-spec list_failed(Level :: task_manager:level()) ->
    {ok, [datastore:document()]} | datastore:generic_error() | no_return().
list_failed(?NON_LEVEL) ->
    {ok, []};

list_failed(?NODE_LEVEL = Level) ->
    Filter = fun
        ('$end_of_table', Acc) ->
            {abort, Acc};
        (#document{value = V} = Doc, Acc) ->
            N = node(),
            case (V#task_pool.node =/= N) orelse task_manager:check_owner(V#task_pool.owner) of
                false ->
                    {next, [Doc | Acc]};
                _ ->
                    {next, Acc}
            end
    end,
    datastore:list(task_to_db_level(Level), ?MODEL_NAME, Filter, []);

list_failed(Level) ->
    Filter = fun
        ('$end_of_table', Acc) ->
            {abort, Acc};
        (#document{value = V} = Doc, Acc) ->
            case task_manager:is_task_alive(V) of
                false ->
                    {next, [Doc | Acc]};
                _ ->
                    {next, Acc}
            end
    end,
    datastore:list(task_to_db_level(Level), ?MODEL_NAME, Filter, []).

%%--------------------------------------------------------------------
%% @doc
%% {@link model_behaviour} callback delete/1.
%% @end
%%--------------------------------------------------------------------
-spec delete(datastore:key()) -> ok | datastore:generic_error().
delete(Key) ->
    datastore:delete(?STORE_LEVEL, ?MODULE, Key).

%%--------------------------------------------------------------------
%% @doc
%% Same as {@link model_behaviour} callback delete/1 but allows
%% choice of task level.
%% @end
%%--------------------------------------------------------------------
-spec delete(Level :: task_manager:level(), datastore:key()) -> ok | datastore:generic_error().
delete(?NON_LEVEL, _Key) ->
    ok;

delete(Level, Key) ->
    datastore:delete(task_to_db_level(Level), ?MODULE, Key).

%%--------------------------------------------------------------------
%% @doc
%% {@link model_behaviour} callback exists/1. 
%% @end
%%--------------------------------------------------------------------
-spec exists(datastore:key()) -> datastore:exists_return().
exists(Key) ->
    ?RESPONSE(datastore:exists(?STORE_LEVEL, ?MODULE, Key)).

%%--------------------------------------------------------------------
%% @doc
%% {@link model_behaviour} callback model_init/0. 
%% @end
%%--------------------------------------------------------------------
-spec model_init() -> model_behaviour:model_config().
model_init() ->
    ?MODEL_CONFIG(task_pool_bucket, [], ?LOCAL_ONLY_LEVEL).

%%--------------------------------------------------------------------
%% @doc
%% {@link model_behaviour} callback 'after'/5. 
%% @end
%%--------------------------------------------------------------------
-spec 'after'(ModelName :: model_behaviour:model_type(),
    Method :: model_behaviour:model_action(),
    Level :: datastore:store_level(), Context :: term(),
    ReturnValue :: term()) -> ok.
'after'(_ModelName, _Method, _Level, _Context, _ReturnValue) ->
    ok.

%%--------------------------------------------------------------------
%% @doc
%% {@link model_behaviour} callback before/4. 
%% @end
%%--------------------------------------------------------------------
-spec before(ModelName :: model_behaviour:model_type(),
    Method :: model_behaviour:model_action(),
    Level :: datastore:store_level(), Context :: term()) ->
    ok | datastore:generic_error().
before(_ModelName, _Method, _Level, _Context) ->
    ok.

%%%===================================================================
%%% API functions
%%%===================================================================

%%--------------------------------------------------------------------
%% @doc
%% Counts alive and failed tasks. Stops counting after specified limit.
%% @end
%%--------------------------------------------------------------------
-spec count_tasks(Level :: task_manager:level(), Type :: atom(), Limit :: non_neg_integer()) ->
    {ok, {Failed :: non_neg_integer(), All :: non_neg_integer()}} | datastore:generic_error() | no_return().
count_tasks(Level, Type, Limit) ->
    Filter = fun
                 ('$end_of_table', {Failed, All}) ->
                     {abort, {Failed, All}};
                 (#document{value = #task_pool{task_type = T, node = N} = V}, {Failed, All}) ->
                     case T of
                         Type ->
                             IsFailed = case Level of
                                 ?NODE_LEVEL ->
                                     MyNode = node(),
                                     (MyNode =:= N) andalso not task_manager:check_owner(V#task_pool.owner);
                                 _ ->
                                     not task_manager:is_task_alive(V)
                             end,
                             NewFailed = case IsFailed of
                                 true ->
                                     Failed + 1;
                                 _ ->
                                     Failed
                             end,
                             NewAll = All + 1,
                             case NewAll >= Limit of
                                 true ->
                                     {abort, {NewFailed, NewAll}};
                                 _ ->
                                     {next, {NewFailed, NewAll}}
                             end;
                         _ ->
                             {next, {Failed, All}}
                     end
             end,
    datastore:list(task_to_db_level(Level), ?MODEL_NAME, Filter, {0,0}).

%%%===================================================================
%%% Internal functions
%%%===================================================================

%%--------------------------------------------------------------------
%% @doc
%% Translates task level to store level.
%% @end
%%--------------------------------------------------------------------
-spec task_to_db_level(Level :: task_manager:level()) -> datastore:store_level().
task_to_db_level(?NODE_LEVEL) ->
    ?LOCAL_ONLY_LEVEL;

task_to_db_level(?CLUSTER_LEVEL) ->
    ?GLOBAL_ONLY_LEVEL;

task_to_db_level(?PERSISTENT_LEVEL) ->
    ?DISK_ONLY_LEVEL.